-----------------------
--- Module Creation ---
-----------------------
local core = LibStub("AceAddon-3.0"):GetAddon("tomQuest2")
local module = core:NewModule("informations", "AceEvent-3.0")

-----------------------
--- Library loading ---
-----------------------
local LC = LibStub("AceLocale-3.0"):GetLocale("tomQuest2-Core")
local LSM = LibStub("LibSharedMedia-3.0")

--------------------------------------------
--- Copy global functions to local scope ---
--------------------------------------------
local format, gsub, type, abs, tonumber = format, gsub, type, math.abs, tonumber
local unpack, tremove, pairs, tconcat = unpack, tremove, pairs, table.concat
local CreateFrame, CreateFont = CreateFrame, CreateFont
local GetQuestLogSelection, SelectQuestLogEntry, QuestLogPushQuest = GetQuestLogSelection, SelectQuestLogEntry, QuestLogPushQuest
local SetAbandonQuest, GetAbandonQuestName, GetAbandonQuestItems = SetAbandonQuest, GetAbandonQuestName, GetAbandonQuestItems
local UIParent, IsShiftKeyDown, IsControlKeyDown, IsAltKeyDown = UIParent, IsShiftKeyDown, IsControlKeyDown, IsAltKeyDown
local GetItemInfo, DressUpItemLink, SetItemRef = GetItemInfo, DressUpItemLink, SetItemRef

-----------------
--- variables ---
-----------------
local _, currentUid, currentPage
local tq2QlFrame, tq2LhFrame, tq2ParentFrame

local stack, tempTable = {}, {}
local repl = {["<"] = "&lt;", [">"] = "&gt;"}

-----------------------
--- Default profile ---
-----------------------
local defaults = {
	profile = {
		showQuestLog = true,
		showLightHeaded = true,
		backDropColor = {0.09, 0.09, 0.19, 0.5},
		borderColor = {0.5, 0.5, 0.5, 1},
		attach = 1,
		showAllCommentsOnOnePage = false,
		qlFont = LSM.DefaultMedia.font,
		qlFontSizeH1 = 14,
		qlFontSizeH2 = 12,
		qlFontSize = 10,
		lhFont = LSM.DefaultMedia.font,
		lhFontSize = 10,
		tomQuest2ParentFrame = {
			width = 640,
			height = 400,
			point = "CENTER",
			x = 0,
			y = 0,
		},
		tomQuest2QlFrame = {
			width = 320,
			height = 400,
			point = "CENTER",
			x = 0,
			y = 0,
		},
		tomQuest2LhFrame = {
			width = 320,
			height = 400,
			point = "CENTER",
			x = 0,
			y = 0,
		},
		color = "ffffd100",
	},
}

----------------------
--- Initialization ---
----------------------
function module:OnInitialize()
	self.db = core.db:RegisterNamespace("informations", defaults, "Default")
	self:SetEnabledState(core.db.profile.modules["informations"])
	core.moduleState["informations"] = core.db.profile.modules["informations"]
end

function module:OnEnable()
	core.RegisterCallback(self, "displayQuestInformations")
	core.RegisterCallback(self, "playerQuestLost", "questLost")
	core.RegisterCallback(self, "partyQuestLost", "questLost")
end

function module:OnDisable()
	core.UnregisterAllCallbacks(self)
end

function module:questLost(event, commHandler, uid, abandoning, title, sender)
	if core.quests[uid] or currentUid ~= uid then return end
	currentUid = nil
	if self.db.profile.showQuestLog and self.db.profile.showLightHeaded and LightHeaded and (self.db.profile.attach == 1 or self.db.profile.attach == 2) then
		if tq2ParentFrame and tq2ParentFrame:IsVisible() then
			tq2ParentFrame:Hide()
			return
		end
	else
		if tq2QlFrame and tq2QlFrame:IsVisible() then
			tq2QlFrame:Hide()
			return
		elseif tq2LhFrame and tq2LhFrame:IsVisible() then
			tq2LhFrame:Hide()
			return
		end
	end
end

local function onLeave(self)
	GameTooltip:Hide()
end

local function onCloseClick(self, button)
	currentUid = nil
	self:GetParent():Hide()
end

local close = CreateFrame("Button", nil, nil, "UIPanelCloseButton")
close:SetScript("OnClick", onCloseClick)

local function onResizeEnter(self)
	GameTooltip:SetOwner(self, "ANCHOR_TOPLEFT")
	GameTooltip:SetText(LC["Click to resize tomQuest2"])
	GameTooltip:Show()
end

local function onResizeMouseDown(self, button)
	self:GetParent():StartSizing()
end

local function onMouseUp(self, button)
	local parentFrame = self:GetParent()
	local name = parentFrame:GetName()
	parentFrame:StopMovingOrSizing()
	if name == "tomQuest2ParentFrame" then
		local width = parentFrame:GetWidth()
		local halfWidth = width/2
		local height = parentFrame:GetHeight()
		tq2QlFrame:SetWidth(halfWidth)
		tq2LhFrame:SetWidth(halfWidth)
		tq2QlFrame:SetHeight(height)
		tq2LhFrame:SetHeight(height)
		tq2QlFrame.text:updateSize()
		tq2LhFrame.text:updateSize()
	else
		parentFrame.text:updateSize()
	end	
	module.db.profile[name].point = "BOTTOMLEFT"
	module.db.profile[name].x = parentFrame:GetLeft()
	module.db.profile[name].y = parentFrame:GetBottom()
	module.db.profile[name].width = width
	module.db.profile[name].height = height
end

local resize = CreateFrame("Button")
resize:SetNormalTexture("Interface\\AddOns\\tomQuest2\\media\\Resize")
resize:SetHighlightTexture("Interface\\AddOns\\tomQuest2\\media\\Resize")
resize:SetHeight(12)
resize:SetWidth(12)
resize:SetScript("OnEnter", onResizeEnter)
resize:SetScript("OnLeave", onLeave)
resize:SetScript("OnMouseDown", onResizeMouseDown)
resize:SetScript("OnMouseUp", onMouseUp)

local function onMovingMouseDown(self, button)
	self:GetParent():StartMoving()
end

local titlereg = CreateFrame("Button")
titlereg:SetHeight(20)
titlereg:SetScript("OnMouseDown", onMovingMouseDown)
titlereg:SetScript("OnMouseUp", onMouseUp)

StaticPopupDialogs["LIGHTHEADED_ABOUT_DIALOG"] = {
	text = "LightHeaded (C) 2007-2008 by Jim Whitehead\n\nThis add-on only provides you with the comments and some basic information on quests. For the full information, be sure to visit http://www.wowhead.com",
	button1 = TEXT(OKAY),
	OnAccept = function()end,
	timeout = 0,
	hideOnEscape = 1
}

local function onIconClick(self, button)
	StaticPopup_Show("LIGHTHEADED_ABOUT_DIALOG")
end

local function onPrevEnter(self)
	GameTooltip:SetOwner(self, "ANCHOR_TOPLEFT")
	GameTooltip:SetText(LC["Previous page"])
	GameTooltip:Show()
end

local function onPrevClick(self, button)
	if currentPage then
		module:updateQuestsLightHeadedInformations(currentUid, currentPage - 1)
	end
end

local function onNextEnter(self)
	GameTooltip:SetOwner(self, "ANCHOR_TOPLEFT")
	GameTooltip:SetText(LC["Next page"])
	GameTooltip:Show()
end

local function onNextClick(self, button)
	if currentPage then
		module:updateQuestsLightHeadedInformations(currentUid, currentPage + 1)
	end
end

local function onDisableClick(self, button)
	if core.quests[currentUid].id then
		local wasSelected = GetQuestLogSelection()
		SelectQuestLogEntry(core.quests[currentUid].id)
		SetAbandonQuest()
		local items = GetAbandonQuestItems()
		if items then
			StaticPopup_Hide("ABANDON_QUEST")
			StaticPopup_Show("ABANDON_QUEST_WITH_ITEMS", GetAbandonQuestName(), items)
		else
			StaticPopup_Hide("ABANDON_QUEST_WITH_ITEMS")
			StaticPopup_Show("ABANDON_QUEST", GetAbandonQuestName())
		end
		SelectQuestLogEntry(wasSelected)
	end
end

local function onShareClick(self, button)
	local wasSelected = GetQuestLogSelection()
	SelectQuestLogEntry(core.quests[currentUid].id)
	QuestLogPushQuest()
	SelectQuestLogEntry(wasSelected)
end

local function onHyperlinkClick(self, link)
	if not link then return end
	if link:match("^lhref:npc") then
		stack[#(stack)+1] = currentPage
		stack[#(stack)+1] = currentUid
		local id, name = link:match("^lhref:npc:(.+):(.+)$")
		id = tonumber(id)
		if LightHeaded:LoadNPCData(id) then
			local tIndex = 1
			tempTable[tIndex] = "|cFF0066FF|Hlhref:back|h[Back]|h|r\n\n\n"..LC["Known locations for "].."|cFFFFFFFF"..name.."|r:"
			for c, z, x, y in LightHeaded:IterateNPCLocs(id) do
				tIndex = tIndex + 1
				tempTable[tIndex] = "|cFF0066FF|Hlhref:zcoord:"..c..":"..z..":"..x..":"..y..":"..name.."|h["..LightHeaded:GetCoordText(x, y).."]|h|r"
			end
			tq2LhFrame.text:mySetText(tconcat(tempTable, "\n", 1, tIndex))
		else
			tq2LhFrame.text:mySetText("|cFF0066FF|Hlhref:back|h[Back]|h|r\n\n\n"..LC["Location unknown"].."|cFFFFFFFF"..name.."|r")
		end
		tq2LhFrame.next:Disable()
		tq2LhFrame.prev:Disable()
	elseif link:match("^lhref:coord") then
		local xt, yt, uid = link:match("^lhref:coord:([^:]+):([^:]+):(.+)")
		local x = tonumber(xt)
		local y = tonumber(yt)
		if not x or not y then
			core:Print(format(LC["The coordinate you clicked on was not valid: %s"], link))
			return
		end
		local note
		if not tonumber(uid) then
			note = uid
		else
			note = "[LH] "..LightHeaded:GetQuestName(uid)
		end
		if TomTom then
			TomTom:AddWaypoint(x, y, note)
		elseif Cartographer_Waypoints and type(Cartographer_Waypoints.AddLHWaypoint) == "function" then
			Cartographer_Waypoints:AddLHWaypoint(nil, nil, x, y, note)
		elseif MN_ThottInterface and type(MN_ThottInterface_Local) == "function" then
			MN_ThottInterface_Local(x, y, note)
		elseif Cartographer_Waypoints then
			core:Print(LC["LightHeaded waypoints are not supported in this version of Cartographer_Waypoints."])
		end
	elseif link:match("^lhref:zcoord") then
		local c, z, x, y, uid = link:match("^lhref:zcoord:([^:]+):([^:]+):([^:]+):([^:]+):(.+)")
		c, z, x, y = tonumber(c), tonumber(z), tonumber(x), tonumber(y)
		local note
		if not tonumber(uid) then
			note = uid
		else
			note = "[LH]"..LightHeaded:GetQuestName(uid)
		end
		if TomTom and type(TomTom.AddZWaypoint) == "function" then
			TomTom:AddZWaypoint(c, z, x, y, note)
		elseif  Cartographer_Waypoints and type(Cartographer_Waypoints.AddLHWaypoint) == "function" then
			Cartographer_Waypoints:AddLHWaypoint(c, z, x, y, note)
		elseif MN_ThottInterface and type(MN_ThottInterface_Legacy) == "function" then
			MN_ThottInterface_Legacy(c, z, x, y, note)
		elseif TomTom then
			core:Print(LC["NPC locations are not supported in this version of TomTom."])
		elseif Cartographer_Waypoints then
			core:Print(LC["LightHeaded waypoints are not supported in this version of Cartographer_Waypoints."])
		end
	elseif link:match("^lhref:external") then
		local url = link:match("^lhref:external:(.+)$")
		ChatFrameEditBox:Show()
		ChatFrameEditBox:SetText(url)
		ChatFrameEditBox:SetFocus()
	elseif link:match("^lhref:quest:") then
		stack[#(stack)+1] = currentPage
		stack[#(stack)+1] = currentUid
		local uid = link:match("^lhref:quest:(.+)$")
		uid = tonumber(uid)
		if self.db.profile.showAllCommentsOnOnePage then
			module:updateQuestsLightHeadedInformations(uid, 0)
			tq2LhFrame.next:Hide()
			tq2LhFrame.prev:Hide()
			tq2LhFrame.page:Hide()
		else
			module:updateQuestsLightHeadedInformations(uid, 1)
		end
	elseif link:match("^lhref:back$") then
		local uid = tremove(stack)
		local page = tremove(stack)
		module:updateQuestsLightHeadedInformations(uid, page)
	elseif link:match("^lhref:item:") or link:match("^item:") then
		if link:match("^lhref:item:") then
			link = link:sub(7)
		end
		local name, chatLink = GetItemInfo(link)
		if name then
			if IsControlKeyDown() then
				DressUpItemLink(link)
			elseif IsShiftKeyDown() and ChatFrameEditBox:IsVisible() then
				ChatFrameEditBox:Insert(chatLink)
			else
				SetItemRef(link)
			end
		end
	elseif not link:match("^lhref:spell:(%d+)") then
		core:Print(format(LC["Unmanaged lhref:%s "], link))
	end
end

local function onHyperlinkEnter(self, link)
	if not link then return end
	if link:match("^lhref:item") then
		GameTooltip:SetOwner(self,"ANCHOR_CURSOR")
		GameTooltip:SetHyperlink(link:sub(7))
	elseif link:match("^item") then
		GameTooltip:SetOwner(self,"ANCHOR_CURSOR")
		GameTooltip:SetHyperlink(link)
	elseif link:match("^lhref:spell:(%d+)") then
		local uid = tonumber(link:match("^lhref:spell:(%d+)"))
		if not uid then return end
		GameTooltip:SetOwner(self,"ANCHOR_CURSOR")
		GameTooltip:SetQuestLogRewardSpell(core.quests[uid].id)
	end
end

-----------------------
--- Frame Functions ---
-----------------------
function module:createLhGUI()
	if tq2LhFrame then return end
	tq2LhFrame = CreateFrame("Frame", "tomQuest2LhFrame", UIParent)
	UISpecialFrames[#UISpecialFrames+1] = "tomQuest2LhFrame"
	tq2LhFrame:SetWidth(640)
	tq2LhFrame:SetHeight(512)
	tq2LhFrame:SetBackdrop({bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", edgeSize = 16, insets = {left = 5, right = 5, top = 5, bottom = 5}, tile = true, tileSize = 16})
	tq2LhFrame:SetBackdropColor(self.db.profile.backDropColor[1], self.db.profile.backDropColor[2], self.db.profile.backDropColor[3], self.db.profile.backDropColor[4])

	tq2LhFrame:EnableMouse()
	tq2LhFrame:SetMovable(1)
	tq2LhFrame:SetResizable(1)
	tq2LhFrame:SetMinResize(300, 300)
	tq2LhFrame:SetWidth(400)
	tq2LhFrame:SetFrameLevel(0)
	tq2LhFrame:SetWidth(325)
	tq2LhFrame:SetHeight(450)

	tq2LhFrame.icon = CreateFrame("Button", nil, tq2LhFrame)
	tq2LhFrame.icon:SetHeight(64)
	tq2LhFrame.icon:SetWidth(64)
	tq2LhFrame.icon:SetNormalTexture("Interface\\AddOns\\LightHeaded\\images\\wh_icon")
	tq2LhFrame.icon:SetHighlightTexture("Interface\\AddOns\\LightHeaded\\images\\wh_icon_hover")
	tq2LhFrame.icon:GetHighlightTexture():SetBlendMode("BLEND")
	tq2LhFrame.icon:SetPoint("BOTTOMRIGHT", -10, -10)
	tq2LhFrame.icon:SetScript("OnClick", onIconClick)

	tq2LhFrame.page = tq2LhFrame:CreateFontString(nil, "ARTWORK")
	tq2LhFrame.page:SetFontObject(GameFontNormalSmall)
	tq2LhFrame.page:SetPoint("BOTTOM", -32, 15)

	tq2LhFrame.prev = CreateFrame("Button", nil, tq2LhFrame)
	tq2LhFrame.prev:SetWidth(21)
	tq2LhFrame.prev:SetHeight(21)
	tq2LhFrame.prev:SetNormalTexture("Interface\\Buttons\\UI-SpellbookIcon-PrevPage-Up")
	tq2LhFrame.prev:SetPushedTexture("Interface\\Buttons\\UI-SpellbookIcon-PrevPage-Down")
	tq2LhFrame.prev:SetDisabledTexture("Interface\\Buttons\\UI-SpellbookIcon-PrevPage-Disabled")
	tq2LhFrame.prev:SetHighlightTexture("Interface\\Buttons\\UI-Common-MouseHilight")
	tq2LhFrame.prev:SetPoint("RIGHT", tq2LhFrame.page, "LEFT", -25, 0)
	tq2LhFrame.prev:SetScript("OnEnter", onPrevEnter)
	tq2LhFrame.prev:SetScript("OnLeave", onLeave)
	tq2LhFrame.prev:SetScript("OnClick", onPrevClick)

	tq2LhFrame.next = CreateFrame("Button", nil, tq2LhFrame)
	tq2LhFrame.next:SetWidth(21)
	tq2LhFrame.next:SetHeight(21)
	tq2LhFrame.next:SetNormalTexture("Interface\\Buttons\\UI-SpellbookIcon-NextPage-Up")
	tq2LhFrame.next:SetPushedTexture("Interface\\Buttons\\UI-SpellbookIcon-NextPage-Down")
	tq2LhFrame.next:SetDisabledTexture("Interface\\Buttons\\UI-SpellbookIcon-NextPage-Disabled")
	tq2LhFrame.next:SetHighlightTexture("Interface\\Buttons\\UI-Common-MouseHilight")
	tq2LhFrame.next:SetPoint("LEFT", tq2LhFrame.page, "RIGHT", 25, 0)
	tq2LhFrame.next:SetScript("OnEnter", onNextLeave)
	tq2LhFrame.next:SetScript("OnLeave", onLeave)
	tq2LhFrame.next:SetScript("OnClick", onNextClick)

	tq2LhFrame.title = tq2LhFrame:CreateFontString(nil, "ARTWORK")
	tq2LhFrame.title:SetPoint("TOP", 0, -4)

	tq2LhFrame.scroll = CreateFrame("ScrollFrame", "tomQuest2LhScrollFrame", tq2LhFrame, "UIPanelScrollFrameTemplate")
	tq2LhFrame.scroll:SetPoint("TOPLEFT", 30, -30)
	tq2LhFrame.scroll:SetPoint("BOTTOMRIGHT", -35, 55)

	tq2LhFrame.text = CreateFrame("SimpleHTML", "tomQuest2LhSimpleHTML", tq2LhFrame.scroll)
	tq2LhFrame.text:SetPoint("TOPLEFT", 0, 0)
	tq2LhFrame.text:SetHeight(300)
	tq2LhFrame.text:SetFontObject(GameFontHighlight)

	local function tq2LhOnUpdateFix(self, elapsed)
		self.text:SetText(self.updateText)
		self.updateText = nil
		self:SetScript("OnUpdate", nil)
	end

	function tq2LhFrame.text:mySetText(msg)
		self.text = msg
		tq2LhFrame.updateText = msg
		tq2LhFrame:SetScript("OnUpdate", tq2LhOnUpdateFix)
	end

	function tq2LhFrame.text:updateSize()
		local lhw = tomQuest2LhFrame:GetWidth()
		local width = lhw - 65
		self:SetWidth(width)
		self:mySetText(self.text)
	end
	tq2LhFrame.text:SetScript("OnHyperlinkClick", onHyperlinkClick)
	tq2LhFrame.text:SetScript("OnHyperlinkEnter", onHyperlinkEnter)
	tq2LhFrame.text:SetScript("OnHyperlinkLeave", onLeave)
	tq2LhFrame.scroll:SetScrollChild(tq2LhFrame.text)
	tq2LhFrame:SetAlpha(1)
	tq2LhFrame:Show()
	self:lockUnlockLhFrame()
end

function module:createQlGUI()
	if tq2QlFrame then return end
	tq2QlFrame = CreateFrame("Frame", "tomQuest2QlFrame", UIParent)
	UISpecialFrames[#UISpecialFrames+1] = "tomQuest2QlFrame"
	tq2QlFrame:SetWidth(640)
	tq2QlFrame:SetHeight(512)
	tq2QlFrame:SetBackdrop({bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", edgeSize = 16, insets = {left = 5, right = 5, top = 5, bottom = 5}, tile = true, tileSize = 16})
	tq2QlFrame:SetBackdropColor(self.db.profile.backDropColor[1], self.db.profile.backDropColor[2], self.db.profile.backDropColor[3], self.db.profile.backDropColor[4])

	tq2QlFrame:EnableMouse()
	tq2QlFrame:SetMovable(1)
	tq2QlFrame:SetResizable(1)
	tq2QlFrame:SetMinResize(300, 300)
	tq2QlFrame:SetWidth(400)
	tq2QlFrame:SetFrameLevel(0)
	tq2QlFrame:SetWidth(325)
	tq2QlFrame:SetHeight(450)

	tq2QlFrame.title = tq2QlFrame:CreateFontString(nil, "ARTWORK")
	tq2QlFrame.title:SetPoint("TOP", 0, -4)

	tq2QlFrame.scroll = CreateFrame("ScrollFrame", "tomQuest2QlScrollFrame", tq2QlFrame, "UIPanelScrollFrameTemplate")
	tq2QlFrame.scroll:SetPoint("TOPLEFT", 30, -30)
	tq2QlFrame.scroll:SetPoint("BOTTOMRIGHT", -35, 55)

	tq2QlFrame.text = CreateFrame("SimpleHTML", "tomQuest2QlSimpleHTML", tq2QlFrame.scroll)
	tq2QlFrame.text:SetPoint("TOPLEFT", 0, 0)
	tq2QlFrame.text:SetHeight(300)
	tq2QlFrame.text:SetFontObject(GameFontHighlight)

	tq2QlFrame.abandon = CreateFrame("Button", "tq2QlAbandonFrame", tq2QlFrame, "UIPanelButtonTemplate")
	tq2QlFrame.abandon:SetWidth(125)
	tq2QlFrame.abandon:SetHeight(21)
	tq2QlFrame.abandon:SetPoint("BOTTOMLEFT", tq2QlFrame, "BOTTOMLEFT", 15, 10)
	tq2QlFrame.abandon:SetText(ABANDON_QUEST)
	tq2QlFrame.abandon:SetScript("OnClick", onDisableClick)
	
	tq2QlFrame.share = CreateFrame("Button", "tq2QlShareFrame", tq2QlFrame, "UIPanelButtonTemplate")
	tq2QlFrame.share:SetWidth(125)
	tq2QlFrame.share:SetHeight(21)
	tq2QlFrame.share:SetPoint("BOTTOMRIGHT", tq2QlFrame, "BOTTOMRIGHT", -15, 10)
	tq2QlFrame.share:SetText(SHARE_QUEST)
	tq2QlFrame.share:SetScript("OnClick", onShareClick)
	
	local function tq2QlOnUpdateFix(self, elapsed)
		self.text:SetText(self.updateText)
		self.updateText = nil
		self:SetScript("OnUpdate", nil)
	end

	function tq2QlFrame.text:mySetText(msg)
		self.text = msg
		tq2QlFrame.updateText = msg
		tq2QlFrame:SetScript("OnUpdate", tq2QlOnUpdateFix)
	end

	function tq2QlFrame.text:updateSize()
		local qlw = tomQuest2QlFrame:GetWidth()
		local width = qlw - 65
		self:SetWidth(width)
		self:mySetText(self.text)
	end
	tq2QlFrame.text:SetScript("OnHyperlinkClick", onHyperlinkClick)
	tq2QlFrame.text:SetScript("OnHyperlinkEnter", onHyperlinkEnter)
	tq2QlFrame.text:SetScript("OnHyperlinkLeave", onLeave)
	tq2QlFrame.scroll:SetScrollChild(tq2QlFrame.text)
	tq2QlFrame:SetAlpha(1)
	tq2QlFrame:Show()
	self:lockUnlockQlFrame()
end

function module:lockUnlockQlFrame()
	tq2QlFrame:ClearAllPoints()
	if self.db.profile.showQuestLog and self.db.profile.showLightHeaded and LightHeaded and (self.db.profile.attach == 1 or self.db.profile.attach == 2) then
		if not tq2ParentFrame then
			tq2ParentFrame = CreateFrame("frame", "tomQuest2ParentFrame", UIParent)
			tq2ParentFrame:SetBackdrop({bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", edgeSize = 16, insets = {left = 5, right = 5, top = 5, bottom = 5}, tile = true, tileSize = 16})
			tq2ParentFrame:SetBackdropColor(self.db.profile.backDropColor[1], self.db.profile.backDropColor[2], self.db.profile.backDropColor[3], self.db.profile.backDropColor[4])
			tq2ParentFrame:SetBackdropBorderColor(self.db.profile.borderColor[1], self.db.profile.borderColor[2], self.db.profile.borderColor[3], self.db.profile.borderColor[4])
			UISpecialFrames[#UISpecialFrames+1] = "tomQuest2ParentFrame"
			close:SetParent(tq2ParentFrame)
			close:SetPoint("TOPRIGHT")
			resize:SetParent(tq2ParentFrame)
			resize:SetPoint("BOTTOMRIGHT", -3, 3)
			titlereg:SetParent(tq2ParentFrame)
			titlereg:SetPoint("TOPLEFT", 5, -5)
			titlereg:SetPoint("TOPRIGHT", 0, 0)
			tq2ParentFrame:EnableMouse()
			tq2ParentFrame:SetMovable(1)
			tq2ParentFrame:SetResizable(1)
			tq2ParentFrame:SetMinResize(600, 300)
			tq2ParentFrame:SetFrameLevel(0)
			tq2ParentFrame:SetWidth(self.db.profile["tomQuest2ParentFrame"].width)
			tq2ParentFrame:SetHeight(self.db.profile["tomQuest2ParentFrame"].height)
		end
		tq2ParentFrame:SetAlpha(1)
		tq2ParentFrame:SetFrameStrata("HIGH")
		tq2QlFrame:SetParent(tomQuest2ParentFrame)
		tq2QlFrame.title:SetFontObject(GameFontHighlight)
		tq2QlFrame.title:SetText("tomQuest2 - Quest Log")
		tq2QlFrame:SetBackdropColor(0, 0, 0, 0)
		tq2QlFrame:SetBackdropBorderColor(0, 0, 0, 0)
		self:restorePosition("tomQuest2ParentFrame")
		tq2QlFrame:SetWidth(self.db.profile["tomQuest2ParentFrame"].width / 2)
		tq2QlFrame:SetHeight(self.db.profile["tomQuest2ParentFrame"].height)
		if self.db.profile.attach == 1 then
			tq2QlFrame:SetPoint("TOPLEFT")
			tq2QlFrame:SetPoint("BOTTOMLEFT")
		elseif self.db.profile.attach == 2 then
			tq2QlFrame:SetPoint("TOPRIGHT")
			tq2QlFrame:SetPoint("BOTTOMRIGHT")
		end
		tq2QlFrame.text:updateSize()
	elseif self.db.profile.showQuestLog then
		tq2QlFrame:SetParent(UIParent)
		tq2QlFrame.title:SetFontObject(GameFontHighlight)
		tq2QlFrame.title:SetText("tomQuest2 - Quest Log")
		close:SetParent(tq2QlFrame)
		close:SetPoint("TOPRIGHT")
		resize:SetParent(tq2QlFrame)
		resize:SetPoint("BOTTOMRIGHT", -3, 3)
		titlereg:SetParent(tq2QlFrame)
		titlereg:SetPoint("TOPLEFT", 5, -5)
		titlereg:SetPoint("TOPRIGHT", 0, 0)
		if tq2ParentFrame then
			tq2ParentFrame:Hide()
		end
		tq2QlFrame:SetAlpha(1)
		tq2QlFrame.close:Show()
		tq2QlFrame.resize:Show()
		tq2QlFrame.titlereg:Show()
		tq2QlFrame:SetBackdropColor(self.db.profile.backDropColor[1], self.db.profile.backDropColor[2], self.db.profile.backDropColor[3], self.db.profile.backDropColor[4])
		tq2QlFrame:SetBackdropBorderColor(self.db.profile.borderColor[1], self.db.profile.borderColor[2], self.db.profile.borderColor[3], self.db.profile.borderColor[4])
		tq2QlFrame:SetFrameStrata("HIGH")
		self:restorePosition("tomQuest2QlFrame")
		tq2QlFrame.text:updateSize()
	elseif not self.db.profile.showQuestLog then
		tq2QlFrame:Hide()
	end

end

function module:lockUnlockLhFrame()
	tq2LhFrame:ClearAllPoints()
	if self.db.profile.showQuestLog and LightHeaded and self.db.profile.showLightHeaded and (self.db.profile.attach == 1 or self.db.profile.attach == 2) then
		tq2LhFrame:SetParent(tq2ParentFrame)
		tq2LhFrame.title:SetFontObject(GameFontHighlight)
		tq2LhFrame.title:SetText("tomQuest2 - LightHeaded")
		tq2LhFrame:SetBackdropColor(0, 0, 0, 0)
		tq2LhFrame:SetBackdropBorderColor(0, 0, 0, 0)
		tq2LhFrame:SetWidth(self.db.profile["tomQuest2ParentFrame"].width / 2)
		tq2LhFrame:SetHeight(self.db.profile["tomQuest2ParentFrame"].height)
		if self.db.profile.attach == 1 then
			tq2LhFrame:SetPoint("TOPRIGHT")
			tq2LhFrame:SetPoint("BOTTOMRIGHT")
		elseif self.db.profile.attach == 2 then
			tq2LhFrame:SetPoint("TOPLEFT")
			tq2LhFrame:SetPoint("BOTTOMLEFT")
		end
		tq2LhFrame.text:updateSize()
	elseif (self.db.profile.showLightHeaded and LightHeaded) then
		tq2LhFrame:SetParent(UIParent)
		tq2LhFrame.title:SetFontObject(GameFontHighlight)
		tq2LhFrame.title:SetText("tomQuest2 - LightHeaded")
		close:SetParent(tq2LhFrame)
		close:SetPoint("TOPRIGHT")
		resize:SetParent(tq2LhFrame)
		resize:SetPoint("BOTTOMRIGHT", -3, 3)
		titlereg:SetParent(tq2LhFrame)
		titlereg:SetPoint("TOPLEFT", 5, -5)
		titlereg:SetPoint("TOPRIGHT", 0, 0)
		if tq2ParentFrame then
			tq2ParentFrame:Hide()
		end
		tq2LhFrame:SetAlpha(1)
		tq2LhFrame.close:Show()
		tq2LhFrame.resize:Show()
		tq2LhFrame.titlereg:Show()
		tq2LhFrame:SetBackdropColor(self.db.profile.backDropColor[1], self.db.profile.backDropColor[2], self.db.profile.backDropColor[3], self.db.profile.backDropColor[4])
		tq2LhFrame:SetBackdropBorderColor(self.db.profile.borderColor[1], self.db.profile.borderColor[2], self.db.profile.borderColor[3], self.db.profile.borderColor[4])
		tq2LhFrame:SetFrameStrata("HIGH")
		self:restorePosition("tomQuest2LhFrame")
		tq2LhFrame.text:updateSize()
	elseif not (self.db.profile.showLightHeaded and LightHeaded) then
		tq2LhFrame:Hide()
	end
end

function module:restorePosition(name)
	local f = _G[name]
	f:ClearAllPoints()
	f:SetPoint(self.db.profile[name].point, self.db.profile[name].x, self.db.profile[name].y)
	f:SetWidth(self.db.profile[name].width)
	f:SetHeight(self.db.profile[name].height)
end

function module:updateQuestsLightHeadedInformations(uid, page)
	if not tq2LhFrame:IsVisible() then return end
	local lhText
	local num = LightHeaded:GetNumQuestComments(uid) or 0
	if #(stack) > 0 then
		lhText = "|cFF0066FF|Hlhref:back|h[Back]|h|r\n"..self:getQuestLightHeadedInformations(uid, page, num)
	else
		lhText = self:getQuestLightHeadedInformations(uid, page, num)
	end
	tq2LhFrame.text:mySetText(lhText)
	if self.db.profile.showAllCommentsOnOnePage then
		tq2LhFrame.page:Hide()
		tq2LhFrame.prev:Hide()
		tq2LhFrame.next:Hide()
	else
		tq2LhFrame.page:SetText(format(LC["Page %s of %s"], page, num + 1))
		tq2LhFrame.page:Show()
		tq2LhFrame.prev:Show()
		tq2LhFrame.next:Show()
		if page == 1 then
			tq2LhFrame.prev:Disable()
		else
			tq2LhFrame.prev:Enable()
		end
		if page == num + 1 then
			tq2LhFrame.next:Disable()
		else
			tq2LhFrame.next:Enable()
		end
	end
	currentUid = uid
	currentPage = page
end


------------------------------
--- LightHeaded & QuestLog ---
------------------------------
function module:clean(text)
	text = gsub(text, "%[ul%]", "")
	text = gsub(text, "%[/ul%]", "")
	text = gsub(text, "%[li%]", "")
	text = gsub(text, "%[/li%]", "")
	text = gsub(text, "%[url=", "")
	text = gsub(text, "%[/url%]", "")
	return text
end

function module:displayQuestInformations(event, uid)
	if not core.quests[uid] then return end
	if self.db.profile.showQuestLog and not tq2QlFrame then
		self:createQlGUI()
	end
	if LightHeaded and self.db.profile.showLightHeaded and not tomQuest2LhFrame then
		self:createLhGUI()
	end
	if self.db.profile.showQuestLog and self.db.profile.showLightHeaded and LightHeaded and (self.db.profile.attach == 1 or self.db.profile.attach == 2) then
		if tq2ParentFrame:IsVisible() and currentUid == uid then
			tq2ParentFrame:Hide()
			currentUid = nil
			return
		end
		tq2ParentFrame:Show()
	else
		if tq2QlFrame and tq2QlFrame:IsVisible() and currentUid == uid then
			tq2QlFrame:Hide()
			currentUid = nil
			return
		elseif tq2LhFrame and tq2LhFrame:IsVisible() and currentUid == uid then
			tq2LhFrame:Hide()
			currentUid = nil
			return
		end
	end
	if self.db.profile.showQuestLog then
		tq2QlFrame:Show()
		currentUid = uid
		local fontN = CreateFont("tq2NFont")
		fontN:SetFont(LSM:Fetch("font", self.db.profile.qlFont), self.db.profile.qlFontSize)
		local fontH1 = CreateFont("tq2H1Font")
		fontH1:SetFont(LSM:Fetch("font", self.db.profile.qlFont), self.db.profile.qlFontSizeH1, "OUTLINE")
		local fontH2 = CreateFont("tq2H2Font")
		fontH2:SetFont(LSM:Fetch("font", self.db.profile.qlFont), self.db.profile.qlFontSizeH2, "OUTLINE")
		tq2QlFrame.title:SetFontObject(fontN)
		tq2QlFrame.text:SetFontObject(fontN)
		tq2QlFrame.text:SetFontObject("h1", fontH1)
		tq2QlFrame.text:SetFontObject("h2", fontH2)

		if core.quests[uid].id then
			tq2QlFrame.abandon:Enable()
			if core.numPartyMembers > 0 and core.quests[uid].sharable then
				tq2QlFrame.share:Enable()
			else
				tq2QlFrame.share:Disable()
			end
		else
			tq2QlFrame.abandon:Disable()
			tq2QlFrame.share:Disable()
		end

		local tIndex = 2
		if core.quests[uid].playersOnQuest > 1 then
			tempTable[1] = "<HTML><BODY><H1>|cff"..core.quests[uid].color..core.quests[uid].tag..core.quests[uid].title.."|r &lt;"..core.quests[uid].playersOnQuest.."&gt;</H1>"
		else
			tempTable[1] = "<HTML><BODY><H1>|cff"..core.quests[uid].color..core.quests[uid].tag..core.quests[uid].title.."|r </H1>"
		end
		if not core.quests[uid].description then
			tempTable[tIndex] = LC["Quests Informations not yet available, please try once synchronization with party is complete"]
			tq2QlFrame.text:mySetText(tconcat(tempTable, "<P><BR/>", 1, tIndex).."</P></BODY></HTML>")
		else
			if core.questsObjectives[uid] then
				tempTable[tIndex] = "<P>"..gsub(core.quests[uid].objective, "[<>]", repl).."</P><P>"
				for objective, qoData in pairs(core.questsObjectives[uid]) do
					tIndex = tIndex + 1
					tempTable[tIndex] = "|cff"..qoData.qoColor..qoData.desc.."|r "..qoData.qoStatus
				end
				tIndex = tIndex + 1
				tempTable[tIndex] = "</P><BR/><H2>"..LC["Description:"].."</H2>"
			else
				tempTable[tIndex] = "<P>"..gsub(core.quests[uid].objective, "[<>]", repl).."</P>"
				tIndex = tIndex + 1
				tempTable[tIndex] = "<H2>"..LC["Description:"].."</H2>"
			end
			tIndex = tIndex + 1
			tempTable[tIndex] = "<P>"..gsub(core.quests[uid].description, "[<>]", repl).."</P>"
			if (core.quests[uid].titleReward ~= "" or core.quests[uid].rewardSpellLink or (core.quests[uid].numRewards + core.quests[uid].numChoices + core.quests[uid].money + core.quests[uid].honor + core.quests[uid].talent) > 0 ) then
				tIndex = tIndex + 1
				tempTable[tIndex] = "<H2>"..LC["Rewards:"].."</H2><P>"
				if core.quests[uid].money > 0 then
					tIndex = tIndex + 1
					tempTable[tIndex] = LC["You will rececive: "]..core:formatMoney(core.quests[uid].money).."<BR/>"
				end
				if core.quests[uid].honor > 0 then
					tIndex = tIndex + 1
					tempTable[tIndex] = LC["Honor points: "]..core.quests[uid].honor.."<BR/>"
				end
				if core.quests[uid].talent > 0 then
					tIndex = tIndex + 1
					tempTable[tIndex] = LC["Talent points: "]..core.quests[uid].talent.."<BR/>"
				end
				if core.quests[uid].titleReward ~= "" then
					tIndex = tIndex + 1
					tempTable[tIndex] = LC["Title: "]..core.quests[uid].titleReward.."<BR/>"
				end
				if core.quests[uid].rewardSpellName ~= "" then
					tIndex = tIndex + 1
					tempTable[tIndex] = "|T"..core.quests[uid].rewardSpellTexture..":32|t |cff71d5ff|Hlhref:spell:"..uid.."|h["..core.quests[uid].rewardSpellName.."]|h|r".."<BR/>"
				end
				if core.quests[uid].numRewards > 0 then
					for i = 1, core.quests[uid].numRewards do
						local numItems = core.quests[uid].itemRewards[i].numItems
						if numItems > 1 then
							if core.quests[uid].itemRewards[i].texture and core.quests[uid].itemRewards[i].link then
								tIndex = tIndex + 1
								tempTable[tIndex] = numItems.." |T"..core.quests[uid].itemRewards[i].texture..":32|t"..core.quests[uid].itemRewards[i].link
							elseif core.quests[uid].itemRewards[i].link then
								tIndex = tIndex + 1
								tempTable[tIndex] = numItems.." "..core.quests[uid].itemRewards[i].link
							end
						else
							if core.quests[uid].itemRewards[i].texture and core.quests[uid].itemRewards[i].link then
								tIndex = tIndex + 1
								tempTable[tIndex] = "|T"..core.quests[uid].itemRewards[i].texture..":32|t"..core.quests[uid].itemRewards[i].link
							elseif core.quests[uid].itemRewards[i].link then
								tIndex = tIndex + 1
								tempTable[tIndex] = core.quests[uid].itemRewards[i].link
							end
						end
					end
				end
				if core.quests[uid].numChoices > 0 then
					tIndex = tIndex + 1
					tempTable[tIndex] = LC["Choose one of the following rewards:"].."<BR/>"
					for i = 1, core.quests[uid].numChoices do
						local numItems = core.quests[uid].itemChoices[i].numItems
						if numItems > 1 then
							if core.quests[uid].itemChoices[i].texture and  core.quests[uid].itemChoices[i].link then
								tIndex = tIndex + 1
								tempTable[tIndex] = numItems.." |T"..core.quests[uid].itemChoices[i].texture..":32|t"..core.quests[uid].itemChoices[i].link
							elseif  core.quests[uid].itemChoices[i].link then
								tIndex = tIndex + 1
								tempTable[tIndex] = numItems.." "..core.quests[uid].itemChoices[i].link
							end
						else
							if core.quests[uid].itemChoices[i].texture and  core.quests[uid].itemChoices[i].link then
								tIndex = tIndex + 1
								tempTable[tIndex] = "|T"..core.quests[uid].itemChoices[i].texture..":32|t"..core.quests[uid].itemChoices[i].link
							elseif  core.quests[uid].itemChoices[i].link then
								tIndex = tIndex + 1
								tempTable[tIndex] = core.quests[uid].itemChoices[i].link
							end
						end
					end
				end
				tq2QlFrame.text:mySetText(tconcat(tempTable, "<BR/>", 1, tIndex).."</P></BODY></HTML>")
			else
				tq2QlFrame.text:mySetText(tconcat(tempTable, "<BR/>", 1, tIndex).."</BODY></HTML>")
			end
		end
	end
	if LightHeaded and self.db.profile.showLightHeaded then
		tq2LhFrame:Show()
		local font = CreateFont("tq2LhFont")
		font:SetFont(LSM:Fetch("font", self.db.profile.lhFont), self.db.profile.lhFontSize)
		tq2LhFrame.title:SetFontObject(font)
		tq2LhFrame.page:SetFontObject(font)
		tq2LhFrame.text:SetFontObject(font)
		currentUid = uid
		local num = LightHeaded:GetNumQuestComments(uid) or 0
		if self.db.profile.showAllCommentsOnOnePage then
			currentPage = 0
			tq2LhFrame.next:Hide()
			tq2LhFrame.prev:Hide()
			tq2LhFrame.page:Hide()
			local lhText = self:getQuestLightHeadedInformations(uid, 0, num)
			if lhText then
				tq2LhFrame.text:mySetText(lhText)
			end
		else
			currentPage = 1
			tq2LhFrame.page:Show()
			local lhText = self:getQuestLightHeadedInformations(uid, 1, num)
			if lhText then
				tq2LhFrame.page:SetText(format(LC["Page %s of %s"], 1, num + 1))
				tq2LhFrame.text:mySetText(lhText)
				if num == 0 then
					tq2LhFrame.prev:Hide()
					tq2LhFrame.next:Hide()
				else
					tq2LhFrame.prev:Show()
					tq2LhFrame.next:Show()
					tq2LhFrame.prev:Disable()
					if currentPage == num + 1 then
						tq2LhFrame.next:Disable()
					else
						tq2LhFrame.next:Enable()
					end
				end
			end
		end
	end
end

function module:getQuestLightHeadedInformations(uid, page, numComments)
	local title = core.quests[uid] and core.quests[uid].title or LightHeaded:GetQuestName(uid)
	if page == 0 or page == 1 then
		local tIndex = 3
		for uid, sharable, level, reqlev, stype, sname, sid, etype, ename, eid, exp, rep, series in LightHeaded:IterateQuestInfo(uid) do
			tempTable[1] = "|c"..self.db.profile.color..LC["Name:|r "]..title
			tempTable[2] = LC["Quest ID:|r"].." |Hlhref:external:http://www.wowhead.com/?quest="..uid.."|h"..uid.."|h"
			if sharable then
				tempTable[3] = LC["Sharable:|r "]..LC["Yes"]
			else
				tempTable[3] = LC["Sharable:|r "]..LC["No"]
			end
			if level ~= "" then
				tIndex = tIndex + 1
				tempTable[tIndex] = LC["Level:|r "]..level
			end
			if reqlev ~= "" then
				tIndex = tIndex + 1
				tempTable[tIndex] = LC["Required Level:|r"]..reqlev
			end
			if stype ~= "" and sname ~= "" then
				tIndex = tIndex + 1
				if stype == "npc" then
					tempTable[tIndex] = LC["Starts:|r "].."|cFF0066FF|Hlhref:npc:"..sid..":"..sname.."|h["..sname.."]|h|r"
				else
					tempTable[tIndex] = LC["Starts:|r "]..sname
				end
			end
			if etype ~= "" and ename ~= "" then
				tIndex = tIndex + 1
				if etype == "npc" then
					tempTable[tIndex] = LC["Ends:|r "].."|cFF0066FF|Hlhref:npc:"..eid..":"..ename.."|h["..ename.."]|h|r"
				else
					tempTable[tIndex] = LC["Ends:|r "]..ename
				end

			end
			if exp ~= "" then
				tIndex = tIndex + 1
				tempTable[tIndex] = LC["Experience:|r "]..exp
			end
			if rep ~= "" then
				tIndex = tIndex + 1
				tempTable[tIndex] = LC["Reputation Gains:|r"]
				for name, value in rep:gmatch("([^\029]+)\029([^\029]+)") do
					tIndex = tIndex + 1
					tempTable[tIndex] = " - "..name.." |r: "..value
				end
			end
			if series ~= "" then
				tIndex = tIndex + 1
				tempTable[tIndex] = LC["Quest Series:|r"]
				for step, id in series:gmatch("([^\029]+)\029([^\029]+)") do
					tIndex = tIndex + 1
					if uid == id then
						tempTable[tIndex] = step.." - "..title.." |r"
					else
						tempTable[tIndex] = step.." - |r|cFF0066FF|Hlhref:quest:"..id.."|h["..LightHeaded:GetQuestName(id).."]|h|r"
					end
				end
			end
		end
		if page == 0 and numComments > 0 then
			tIndex = tIndex + 1
			tempTable[tIndex] = LC["Comments for:|r "]..title
			for page = 1, numComments do
				tIndex = tIndex + 1
				tempTable[tIndex] = self:clean(LightHeaded:GetCommentText(uid, page))
			end
			return tconcat(tempTable, "\n|c"..self.db.profile.color, 1, tIndex - numComments).."\n"..tconcat(tempTable, "\n", tIndex - numComments + 1, tIndex)
		else
			return tconcat(tempTable, "\n|c"..self.db.profile.color, 1, tIndex)
		end
	elseif page > 1 then
		return "|c"..self.db.profile.color..LC["Comments for:|r "]..title.."\n"..self:clean(LightHeaded:GetCommentText(uid, page - 1))
	end
end