-----------------------
--- Module Creation ---
-----------------------
local core = LibStub("AceAddon-3.0"):GetAddon("tomQuest2")
local module = core:NewModule("syncParty", "AceEvent-3.0", "AceComm-3.0", "AceSerializer-3.0")

--------------------------------------------
--- Copy global functions to local scope ---
--------------------------------------------
local pairs, type, format, time, UnitName, GetNumPartyMembers, len, sub, byte, UnitXP, UnitXPMax, floor = pairs, type, string.format, time, UnitName, GetNumPartyMembers, string.len, string.sub, string.byte, UnitXP, UnitXPMax, math.floor

-----------------
--- variables ---
-----------------
local addonPrefix = "tomQ2"
local version = 252
local firstCompatibleVersion = 252
local lastIncompatibleVersion = 251
local syncParty = false
local sendXPMax = false
local MAX_PLAYER_LEVEL = MAX_PLAYER_LEVEL
local partyZone = core.LC["PARTYQUESTSZONE"]
local questTimer_sendTime = {}
local questTimer_sender = {}
local numTimers = 0
local pId = {true, true, true, true}
for i=1, 4 do
	pId[i] = format("party%d", i)
end

local function normalizeColorComponent(component)
	if component > 1 then return 1 elseif component < 0 then return 0 else return component end
end

local function getColorFromCompletionHEX(percent)
	if percent <= 0 then
		return format("%02x%02x%02x", 255*core.db.profile.quests.rj0, 255*core.db.profile.quests.gj0, 255*core.db.profile.quests.bj0)
	elseif percent >= 1 then
		return format("%02x%02x%02x", 255*core.db.profile.quests.rj3, 255*core.db.profile.quests.gj3, 255*core.db.profile.quests.bj3)
	elseif percent <= 0.5 then
		return format("%02x%02x%02x", 255*normalizeColorComponent(core.db.profile.quests.ri1*percent + core.db.profile.quests.rj1), 255*normalizeColorComponent(core.db.profile.quests.gi1*percent + core.db.profile.quests.gj1), 255*normalizeColorComponent(core.db.profile.quests.bi1*percent + core.db.profile.quests.bj1))
	else
		return format("%02x%02x%02x", 255*normalizeColorComponent(core.db.profile.quests.ri2*percent + core.db.profile.quests.rj2), 255*normalizeColorComponent(core.db.profile.quests.gi2*percent + core.db.profile.quests.gj2), 255*normalizeColorComponent(core.db.profile.quests.bi2*percent + core.db.profile.quests.bj2))
	end
end

local function short(value)
	if value >= 10000000 then
		return format("%.1fm", value / 1000000)
	elseif value >= 1000000 then
		return format("%.2fm", value / 1000000)
	elseif value >= 100000 then
		return format("%.0fk", value / 1000)
	elseif value >= 10000 then
		return format("%.1fk", value / 1000)
	else
		return format("%d", value)
	end
end

-------------------
--- timed Queue ---
-------------------
local function timedQueue(self, elapsed)
	self.duration = self.duration + elapsed
	if self.duration >= 10 then
		self.currentTime = self.currentTime + self.duration
		self.duration = 0
		for uid, time in pairs(questTimer_sendTime) do
			if self.currentTime <= time then
				if core.quests[uid] and core.party[questTimer_sender[uid]] and not core.quests[uid].extras then
					module:SendCommMessage(addonPrefix, module:Serialize("e", uid), "WHISPER", questTimer_sender[uid], "BULK")
				end
				questTimer_sendTime[uid] = nil
				questTimer_sender[uid] = nil
				numTimers = numTimers - 1
			end
		end
		if numTimers == 0 then
			self:Hide()
		end
	end
end
local timedMessageQueue = CreateFrame("frame", "tomQuest2TimedMessageQueue")
timedMessageQueue.duration = 0
timedMessageQueue:SetScript("OnUpdate", timedQueue)
timedMessageQueue:Hide()

local function timedXPQueue(self, elapsed)
	self.duration = self.duration + elapsed
	if self.duration >= 60 then
		self.duration = 0
		self:Hide()
		if sendXPMax then
			module:SendCommMessage(addonPrefix, module:Serialize("u", UnitXP("player"), UnitXPMax("player")), "PARTY", nil, "BULK")
			sendXPMax = false
		else
			module:SendCommMessage(addonPrefix, module:Serialize("x", UnitXP("player")), "PARTY", nil, "BULK")
		end
	end
end
local timedXPQueueFrame = CreateFrame("frame", "tomQuest2TimedXPQueue")
timedXPQueueFrame.duration = 0
timedXPQueueFrame:SetScript("OnUpdate", timedXPQueue)
timedXPQueueFrame:Hide()

function module:OnInitialize()
	self:SetEnabledState(core:getModuleEnabled("syncParty"))
	core.moduleState["syncParty"] = core:getModuleEnabled("syncParty")
end

function module:OnEnable()
	self:RegisterComm(addonPrefix, "commHandler")
	core.RegisterCallback(self, "partyChanged")
	core.RegisterCallback(self, "playerJoinedParty")
	core.RegisterCallback(self, "playerLeftParty")
end

function module:OnDisable()
	core.UnregisterAllCallbacks(self)
end

function module:playerJoinedParty()
	core.RegisterCallback(self, "addPlayerQuest", "sendMessage")
	core.RegisterCallback(self, "addPlayerQuestObjective", "sendMessage")
	core.RegisterCallback(self, "playerQuestAccepted", "sendMessage")
	core.RegisterCallback(self, "playerQuestObjectiveProgress", "sendMessage")
	core.RegisterCallback(self, "playerQuestComplete", "sendMessage")
	core.RegisterCallback(self, "playerQuestUnComplete", "sendMessage")
	core.RegisterCallback(self, "playerQuestFailed", "sendMessage")
	core.RegisterCallback(self, "playerQuestLost", "sendMessage")
	core.RegisterCallback(self, "endSync", "sendMessage")
	if core.playerLevel ~= MAX_PLAYER_LEVEL then
		self:RegisterEvent("PLAYER_XP_UPDATE")
		self:RegisterEvent("PLAYER_LEVEL_UP")
	end
end

function module:playerLeftParty()
	core.UnregisterCallback(self, "addPlayerQuest")
	core.UnregisterCallback(self, "addPlayerQuestObjective")
	core.UnregisterCallback(self, "playerQuestAccepted")
	core.UnregisterCallback(self, "playerQuestObjectiveProgress")
	core.UnregisterCallback(self, "playerQuestComplete")
	core.UnregisterCallback(self, "playerQuestUnComplete")
	core.UnregisterCallback(self, "playerQuestFailed")
	core.UnregisterCallback(self, "playerQuestLost")
	core.UnregisterCallback(self, "endSync")
	self:UnregisterEvent("PLAYER_XP_UPDATE")
	self:UnregisterEvent("PLAYER_LEVEL_UP")
end

local function utf8sub(name)
	local bytes = len(name)
	if bytes <= 3 then
		return name
	else
		local len, pos = 0, 1
		while(pos <= bytes) do
			len = len + 1
			local c = byte(name, pos)
			if (c > 0 and c <= 127) then
				pos = pos + 1
			elseif (c >= 192 and c <= 223) then
				pos = pos + 2
			elseif (c >= 224 and c <= 239) then
				pos = pos + 3
			elseif (c >= 240 and c <= 247) then
				pos = pos + 4
			end
			if (len == 3) then break end
		end
		return sub(name, 1, pos - 1)
	end
end

function module:partyChanged()
	local timeStamp = time()
	local partySync = 0
	if core.numPartyMembers > 0 then
		for i=1, core.numPartyMembers do
			local partyName, partyRealm = UnitName(pId[i])
			if partyRealm and partyRealm ~= "" then
				partyName = partyName.."-"..partyRealm
			end
			if not core.party[partyName] then
				core.party[partyName] = {}
				core.party[partyName].status = 0
				core.party[partyName].subName = utf8sub(partyName)
				self:SendCommMessage(addonPrefix, self:Serialize("v", version, firstCompatibleVersion, core.playerClass, UnitXP("player"), UnitXPMax("player")), "WHISPER", partyName, "BULK")
			elseif not core.party[partyName].timeStamp then
				self:SendCommMessage(addonPrefix, self:Serialize("v", version, firstCompatibleVersion, core.playerClass, UnitXP("player"), UnitXPMax("player")), "WHISPER", partyName, "BULK")
			end
			core.party[partyName].timeStamp = timeStamp
			partySync = partySync + core.party[partyName].status
		end
	end
	for partyName, partyData in pairs(core.party) do
		if partyData.timeStamp ~= timeStamp then
			for uid, questData in pairs(core.quests) do
				if questData.complete[partyName] then
					core:removePartyFromQuest(partyName, uid)
				end
			end
			core.party[partyName] = nil
		end
	end
	if partySync == 0 then
		syncParty = false
	else
		syncParty = true
	end
end


function module:PLAYER_XP_UPDATE()
	timedXPQueueFrame:Show()
end

function module:PLAYER_LEVEL_UP(event, newLevel)
	if newLevel == MAX_PLAYER_LEVEL then
		self:UnregisterEvent("PLAYER_XP_UPDATE")
		self:UnregisterEvent("PLAYER_LEVEL_UP")
	else
		timedXPQueueFrame:Show()
		self.duration = 50
		sendXPMax = true
	end
end
-------------
--- INPUT ---
-------------
local comm = {
	x = function(sender, currXP)
		if not core.party[sender] then return end
		core.party[sender].currXP = currXP
		core.party[sender].fmtXP = "|cff"..getColorFromCompletionHEX(currXP/core.party[sender].maxXP)..short(currXP).." / "..short(core.party[sender].maxXP).." ("..(floor(currXP/core.party[sender].maxXP*10000)/100).."%)|r"
	end,
	u = function(sender, currXP, maxXP)
		if not core.party[sender] then return end
		core.party[sender].currXP = currXP
		core.party[sender].maxXP = maxXP
		core.party[sender].fmtXP = "|cff"..getColorFromCompletionHEX(currXP/maxXP)..short(currXP).." / "..short(maxXP).." ("..(floor(currXP/maxXP*10000)/100).."%)|r"
	end,
	v = function(sender, senderVersion, senderFirstCompatibleVersion, senderClass, currXP, maxXP)
		if senderVersion <= lastIncompatibleVersion then
			core:Print(format(core.LC["The tomQuest2 release (%d) used by %s is too old to synchronise with your release (%d) (min %d)"], senderVersion, sender, version, firstCompatibleVersion))
			return
		elseif version < senderFirstCompatibleVersion then
			core:Print(format(core.LC["Your tomQuest2 release (%d) is too old to synchorise with release (%d) of %s (min %d)"], version, senderVersion, sender, firstCompatibleVersion))
			return
		elseif version < senderVersion then
			core:Print(format(core.LC["New release available (%d) ask %s where to find it"], senderVersion, sender))
		end
		if not core.party[sender] then
			core.party[sender] = {}
			core.party[sender].subName = utf8sub(sender)
		end
		core.party[sender].status = 1
		core.party[sender].class = senderClass
		core.party[sender].color = core.classColors[senderClass]
		core.party[sender].currXP = currXP
		core.party[sender].maxXP = maxXP
		core.party[sender].fmtXP = "|cff"..getColorFromCompletionHEX(currXP/maxXP)..short(currXP).." / "..short(maxXP).." ("..(floor(currXP/maxXP*10000)/100).."%)|r"
		core.party[sender].transfer = true
		syncParty = true
		if core.numQuests > 0 then
			module:SendCommMessage(addonPrefix, module:Serialize("w", version, firstCompatibleVersion, core.playerClass, UnitXP("player"), UnitXPMax("player")), "WHISPER", sender, "BULK")
			for uid, qData in pairs(core.quests) do
				if qData.complete[core.playerName] then
					module:SendCommMessage(addonPrefix, module:Serialize("q", uid, qData.title, qData.level, qData.shortTag, qData.group, qData.daily, qData.numObjectives, qData.complete[core.playerName], qData.zone), "WHISPER", sender, "BULK")
					local begun = 0
					if core.questsObjectives[uid] then
						for objective, qoData in pairs(core.questsObjectives[uid]) do
							if qoData.got[core.playerName] then
								module:SendCommMessage(addonPrefix, module:Serialize("o", uid, objective, qoData.desc, qoData.qType, qoData.got[core.playerName], qoData.need, qoData.nGot[core.playerName], qoData.nNeed), "WHISPER", sender, "BULK")
								if qoData.nGot[core.playerName] ~= 0 then begun = 1 end
							end
						end
					end
					module:SendCommMessage(addonPrefix, module:Serialize("r", uid, begun), "WHISPER", sender, "BULK")
				end
			end
			module:SendCommMessage(addonPrefix, module:Serialize("d"), "WHISPER", sender, "BULK")
		end
	end,
	w = function(sender, senderVersion, senderFirstCompatibleVersion, senderClass, currXP, maxXP)
		if senderVersion <= lastIncompatibleVersion then
			core:Print(format(core.LC["The tomQuest2 release (%d) used by %s is too old to synchronise with your release (%d) (min %d)"], senderVersion, sender, version, firstCompatibleVersion))
			return
		elseif version < senderFirstCompatibleVersion then
			core:Print(format(core.LC["Your tomQuest2 release (%d) is too old to synchorise with release (%d) of %s (min %d)"], version, senderVersion, sender, firstCompatibleVersion))
			return
		elseif version < senderVersion then
			core:Print(format(core.LC["New release available (%d) ask %s where to find it"], senderVersion, sender))
		end
		if not core.party[sender] then
			core.party[sender] = {}
			core.party[sender].subName = utf8sub(sender)
		end
		core.party[sender].status = 1
		core.party[sender].class = senderClass
		core.party[sender].color = core.classColors[senderClass]
		core.party[sender].currXP = currXP
		core.party[sender].maxXP = maxXP
		core.party[sender].fmtXP = "|cff"..getColorFromCompletionHEX(currXP/maxXP)..short(currXP).." / "..short(maxXP).." ("..(floor(currXP/maxXP*10000)/100).."%)|r"
		core.party[sender].transfer = true
		syncParty = true
	end,
	d = function(sender)
		if core.party[sender].status == 1 then
			core.party[sender].transfer = nil
			core.cbh:Fire("updateQuestsTracker")
		end
	end,
	q = function(sender, uid, title, level, tag, group, daily, numObjectives, completed, zone)
		if not core.party[sender] or core.party[sender].status == 0 then return end
		if not core.quests[uid] then
			if core.db.profile.quests.partyHeader then
				if not core.zones[partyZone] then
					core.addZone(partyZone)
					core.zonesSort()
				end
				core.zones[partyZone].numQuests = core.zones[partyZone].numQuests + 1
				core.questsByZone[partyZone][uid] = true
			else
				if not core.zones[zone] then
					core.addZone(zone)
					core.zonesSort()
				end
				core.zones[zone].numQuests = core.zones[zone].numQuests + 1
				core.questsByZone[zone][uid] = true
			end
			core.addQuest(uid, title, level, tag, group, daily, numObjectives, zone)
			core.quests[uid].complete[sender] = completed
			core.quests[uid].playersOnQuest = core.quests[uid].playersOnQuest + 1
			core:questsSort()
		else
			core.quests[uid].complete[sender] = completed
			core.quests[uid].playersOnQuest = core.quests[uid].playersOnQuest + 1
		end
		core:computeQuestStatus(uid)
		if not core.party[sender].transfer and not core.quests[uid].extras and (not core.quests[uid].objective or not core.quests[uid].description) and not questTimer_sendTime[uid] then
			local currentTime = time()
			questTimer_sendTime[uid] = currentTime + 60
			questTimer_sender[uid] = sender
			timedMessageQueue.duration = 0
			timedMessageQueue.currentTime = time()
			numTimers = numTimers + 1
			timedMessageQueue:Show()
		elseif core.party[sender].transfer and not core.quests[uid].extras and (not core.quests[uid].objective or not core.quests[uid].description) then
			module:SendCommMessage(addonPrefix, module:Serialize("e", uid), "WHISPER", sender, "BULK")
		end
	end,
	o = function(sender, uid, objective, desc, qType, got, need, nGot, nNeed)
		if not core.party[sender] or core.party[sender].status == 0 or not core.quests[uid] or not core.questsObjectives[uid] then return end
		if not core.questsObjectives[uid][objective] then
			core.addQuestObjective(uid, objective, desc, qType, need, nNeed)
		end
		core.questsObjectives[uid][objective].got[sender] = got
		core.questsObjectives[uid][objective].nGot[sender] = nGot
		core:computeQuestObjectiveStatus(uid, objective)
	end,
	p = function(sender, uid, objective, got, nGot)
		if not core.party[sender] or core.party[sender].status == 0 or not core.quests[uid] or not core.questsObjectives[uid] or not core.questsObjectives[uid][objective] or not core.questsObjectives[uid][objective].got then return end
		core.questsObjectives[uid][objective].got[sender] = got
		core.questsObjectives[uid][objective].nGot[sender] = nGot
		core:computeQuestObjectiveStatus(uid, objective)
		core.cbh:Fire("partyQuestObjectiveProgress", nil, uid, objective, got, nGot, sender)
	end,
	r = function(sender, uid, numObjectives, started)
		if not core.party[sender] or core.party[sender].status == 0 or not core.quests[uid] or not core.quests[uid].complete[sender] then return end
		core.cbh:Fire("partyQuestAccepted", sender, uid, started, core.party[sender].transfer)
	end,
	e = function(sender, uid)
		if not core.party[sender] or core.party[sender].status == 0 or not core.quests[uid] or not core.quests[uid].complete[core.playerName] then return end
		if core.quests[uid].objective then
			module:SendCommMessage(addonPrefix, module:Serialize("n", uid, core.quests[uid].objective), "WHISPER", sender, "BULK")
		end
		if core.quests[uid].description then
			module:SendCommMessage(addonPrefix, module:Serialize("t", uid, core.quests[uid].description), "WHISPER", sender, "BULK")
		end
		module:SendCommMessage(addonPrefix, module:Serialize("g", uid, core.quests[uid].money or 0, core.quests[uid].honor or 0, core.quests[uid].talent or 0, core.quests[uid].titleReward or "", core.quests[uid].numRewards or 0, core.quests[uid].numChoices or 0, core.quests[uid].rewardSpellName or "", core.quests[uid].rewardSpellTexture or ""), "WHISPER", sender, "BULK")
		if core.quests[uid].numRewards and core.quests[uid].numRewards > 0 then
			for i, irData in pairs (core.quests[uid].itemRewards) do
				module:SendCommMessage(addonPrefix, module:Serialize("i", uid, i, irData.name, irData.texture, irData.numItems, irData.quality, irData.isUsable, irData.link), "WHISPER", sender, "BULK")
			end
		end
		if core.quests[uid].numChoices and core.quests[uid].numChoices > 0 then
			for i, icData in pairs (core.quests[uid].itemChoices) do
				module:SendCommMessage(addonPrefix, module:Serialize("j", uid, i, icData.name, icData.texture, icData.numItems, icData.quality, icData.isUsable, icData.link), "WHISPER", sender, "BULK")
			end
		end
	end,
	n = function(sender, uid, objective)
		if not core.party[sender] or core.party[sender].status == 0 or not core.quests[uid] or core.quests[uid].objective then return end
		core.quests[uid].objective = objective
		core.cbh:Fire("questObjectiveAdded", uid)
	end,
	t = function(sender, uid, description)
		if not core.party[sender] or core.party[sender].status == 0 or not core.quests[uid] or core.quests[uid].description then return end
		core.quests[uid].description = description
		core.cbh:Fire("questDescriptionAdded", uid)
	end,
	g = function(sender, uid, money, honor, talent, titleReward, numRewards, numChoices, rewardSpellName, rewardSpellTexture)
		if not core.party[sender] or core.party[sender].status == 0 or not core.quests[uid] then return end
		core.quests[uid].money = money
		core.quests[uid].honor = honor
		core.quests[uid].talent = talent
		core.quests[uid].titleReward = titleReward
		core.quests[uid].numRewards = numRewards
		core.quests[uid].numChoices = numChoices
		core.quests[uid].rewardSpellName = rewardSpellName
		core.quests[uid].rewardSpellTexture = rewardSpellTexture
	end,
	i = function(sender, uid, i, name, texture, numItems, quality, isUsable, link)
		if not core.party[sender] or core.party[sender].status == 0 or not core.quests[uid] then return end
		core.quests[uid].itemRewards = core.quests[uid].itemRewards or {}
		if not core.quests[uid].itemRewards[i] then
			core.quests[uid].itemRewards[i] = {}
			core.quests[uid].itemRewards[i].name = name
			core.quests[uid].itemRewards[i].texture = texture
			core.quests[uid].itemRewards[i].numItems = numItems
			core.quests[uid].itemRewards[i].quality = quality
			core.quests[uid].itemRewards[i].isUsable = isUsable
			core.quests[uid].itemRewards[i].link = link
		end
	end,
	j = function(sender, uid, i, name, texture, numItems, quality, isUsable, link)
		if not core.party[sender] or core.party[sender].status == 0 or not core.quests[uid] then return end
		core.quests[uid].itemChoices = core.quests[uid].itemChoices or {}
		if not core.quests[uid].itemChoices[i] then
			core.quests[uid].itemChoices[i] = {}
			core.quests[uid].itemChoices[i].name = name
			core.quests[uid].itemChoices[i].texture = texture
			core.quests[uid].itemChoices[i].numItems = numItems
			core.quests[uid].itemChoices[i].quality = quality
			core.quests[uid].itemChoices[i].isUsable = isUsable
			core.quests[uid].itemChoices[i].link = link
		end
	end,
	c = function(sender, uid)
		if not core.party[sender] or core.party[sender].status == 0 or not core.quests[uid] or not core.quests[uid].complete[sender] then return end
		core.quests[uid].complete[sender] = 1
		core:computeQuestStatus(uid)
		core.cbh:Fire("partyQuestComplete", nil, uid, sender)
	end,
	a = function(sender, uid)
		if not core.party[sender] or core.party[sender].status == 0 or not core.quests[uid] or not core.quests[uid].complete[sender] then return end
		core.quests[uid].complete[sender] = 0
		core:computeQuestStatus(uid)
		core.cbh:Fire("partyQuestUnComplete", nil, uid, sender)
	end,
	f = function(sender, uid)
		if not core.party[sender] or core.party[sender].status == 0 or not core.quests[uid] or not core.quests[uid].complete[sender] then return end
		core.quests[uid].complete[sender] = -1
		core:computeQuestStatus(uid)
		core.cbh:Fire("partyQuestFailed", nil, uid, sender)
	end,
	l = function(sender, uid, abandoning)
		if not core.party[sender] or core.party[sender].status == 0 or not core.quests[uid] or not core.quests[uid].complete[sender] then return end
		if questTimer_sendTime[uid] then
			questTimer_sendTime[uid] = nil
			questTimer_sender[uid] = nil
			numTimers = numTimers - 1
			if numTimers == 0 then
				timedMessageQueue:Hide()
			end
		end
		local title = core.quests[uid].title
		core:removePartyFromQuest(sender, uid, abandoning)
		core.cbh:Fire("partyQuestLost", nil, uid, abandoning, title, sender)
	end,
}

function module:commHandler(prefix, message, distribution, sender)
	if sender ~= core.playerName then
		self:onCommand(sender, self:Deserialize(message))
	end
end

function module:onCommand(sender, success, commHandler, ...)
	if not success then
		return
	elseif comm[commHandler] then
		comm[commHandler](sender, ...)
	end
end
--------------
--- OUTPUT ---
--------------
function module:sendMessage(event, commHandler, uid, ...)
	if commHandler == "q" and questTimer_sendTime[uid] then
		questTimer_sendTime[uid] = nil
		questTimer_sender[uid] = nil
		numTimers = numTimers - 1
		if numTimers == 0 then
			timedMessageQueue:Hide()
		end
	end
	if syncParty then
		self:SendCommMessage(addonPrefix, self:Serialize(commHandler, uid, ...), "PARTY", nil, "BULK")
	end
end